#include "TPhysics.hpp"
#include "bench_tools.hpp"

#include <algorithm>
#include <cmath>
#include <exception>
#include <iostream>
#include <iterator>
#include <limits>
#include <numeric>
#include <random>
#include <string>
#include <vector>


int main(int argc, char *argv[]) {
    // Check CLI arguments
    size_t amount, iterations;
    switch (argc) {
        case 1:
            std::cout << "Usage: " << argv[0] << " <amount of bodies> <number of iterations>" << std::endl;
            return 0;
        case 3: {
            amount = std::stoll(argv[1]);
            // Storage is particle-dominated, and particles have a 3-dimensional
            // position, speed, force, plus an associated pointer.
            constexpr size_t PER_PARTICLE_SIZE =
                sizeof(std::unique_ptr<TParticle>)
                    + 3 * (sizeof(TVector) + 3 * sizeof(TScalar));
            if (amount > (2 * 1024 * 1024 * 1024ULL / PER_PARTICLE_SIZE)) {
                std::cout << "2 GB of working set ought to be enough !" << std::endl;
                return 1;
            }
            iterations = std::stoll(argv[2]);
            break;
        }
        default:
            std::cout << "Invalid number of arguments (expected 2, got " << argc - 1 << ')' << std::endl;
            return 1;
    }

    // Set up simulation, iterate repeatedly
    TRandomEngine rng;
    TPhysics simulation(rng, amount);
    for (size_t i = 0; i < iterations; ++i) simulation.step();
    assumeAccessed(simulation);

    // ...and we're done
    return 0;
}


TVector TVector::unit(size_t unit_dim, size_t num_dims) {
    checkDimInRange(unit_dim, num_dims);
    std::vector<TScalar> coords;
    coords.reserve(num_dims);
    for (size_t dim = 0; dim < num_dims; ++dim) {
        coords.push_back(static_cast<TScalar>(dim == unit_dim));
    }
    return TVector(coords);
}


TScalar TVector::operator[](size_t dim) const {
    checkDimInRange(dim);
    return m_coords[dim];
}


TScalar& TVector::operator[](size_t dim) {
    checkDimInRange(dim);
    return m_coords[dim];
}


TScalar TVector::dot(const TVector& other) const {
    checkDimsMatch(other);
    return std::transform_reduce(m_coords.cbegin(), m_coords.cend(),
                                 m_coords.cbegin(),
                                 0.0);
}


void TVector::checkDimInRange(size_t dim, size_t num_dims) {
    if (dim >= num_dims) {
        throw std::runtime_error("Bad coordinate dimension");
    }
}


void TVector::checkDimsMatch(const TVector& other) const {
    if (dims() != other.dims()) {
        throw std::runtime_error("Vector dimensions don't match");
    }
}


TVector TVector::scalarOp(std::function<TScalar(TScalar)>&& op) const noexcept {
    TVector result(*this);
    for (size_t dim = 0; dim < dims(); ++dim) {
        result.m_coords[dim] = op(m_coords[dim]);
    }
    return result;
}


TVector TVector::vectorOp(std::function<TScalar(TScalar, TScalar)>&& op,
                          const TVector& other) const {
    checkDimsMatch(other);
    TVector result(*this);
    for (size_t dim = 0; dim < dims(); ++dim) {
        result.m_coords[dim] = op(m_coords[dim], other.m_coords[dim]);
    }
    return result;
}


TBox::TBox(const TVector& position, const TVector& extent)
    : m_start(position - extent/2)
    , m_end(position + extent/2)
{
    if (!extent.isPositive()) {
        throw std::runtime_error("Box extent must be positive");
    }
}


std::vector<TVector> TBox::normalsToInside() const noexcept {
    std::vector<TVector> result;
    result.reserve(2 * dims());
    for (size_t dim = 0; dim < dims(); ++dim) {
        result.push_back(TVector::unit(dim, dims()));
        result.push_back(-TVector::unit(dim, dims()));
    }
    return result;
}


std::vector<TVector> TBox::projectOnSides(const TVector& position) const {
    if (dims() != position.dims()) {
        throw std::runtime_error("Vector dimension doesn't match box dimension");
    }

    std::vector<TVector> result;
    result.reserve(2 * dims());
    for (size_t dim = 0; dim < dims(); ++dim) {
        TVector projection = position;
        projection[dim] = m_start[dim];
        result.push_back(projection);
        projection[dim] = m_end[dim];
        result.push_back(projection);
    }
    return result;
}


std::vector<TVector> TBox::pick(TRandomEngine& rng,
                                size_t amount) const noexcept {
    // Prepare to generate random coordinates falling strictly within the box
    std::vector<TScalar> coords(dims());
    std::vector<std::uniform_real_distribution<TScalar>> coord_dists;
    coord_dists.reserve(dims());
    for (size_t dim = 0; dim < dims(); ++dim) {
        coord_dists.emplace_back(
            std::nextafter(m_start[dim], std::numeric_limits<TScalar>::max()),
            // This nextafter should not be necessary, but is due to a common
            // spec violation in std::uniform_real_distribution implementations.
            std::nextafter(m_end[dim], std::numeric_limits<TScalar>::lowest())
        );
    }

    // Generate random positions within the box
    std::vector<TVector> results;
    results.reserve(amount);
    for (size_t pos = 1; pos < amount; ++pos) {
        for (size_t dim = 0; dim < dims(); ++dim) {
            coords[dim] = coord_dists[dim](rng);
        }
        results.push_back(TVector(coords));
    }
    return results;
}


std::vector<std::unique_ptr<TParticle>> TParticle::generate(
    TRandomEngine& rng,
    const TBox& box,
    size_t amount
) noexcept {
    // Generate a bunch of particle positions inside the box
    std::vector<TVector> positions = box.pick(rng, amount);

    // Generate particles of random type at those positions
    std::uniform_int_distribution<uint8_t> type_dist(0, 2);
    std::vector<std::unique_ptr<TParticle>> particles;
    particles.reserve(amount);
    std::transform(
        positions.cbegin(), positions.cend(),
        std::back_inserter(particles),
        [&](const TVector& position) -> std::unique_ptr<TParticle> {
            switch (type_dist(rng)) {
                case 0:
                    return std::make_unique<TNeutron>(position);
                case 1:
                    return std::make_unique<TElectron>(position);
                case 2:
                    return std::make_unique<TProton>(position);
                default:
                    throw std::runtime_error("Unreachable case");
            }
        }
    );

    // If some particles don't fully fit in the box (because of radius), just
    // re-roll their position until they do. For a sanely shaped box which is
    // much larger than particle radius this loop will converge very quickly.
    while(true) {
        // Put particles that are outside the box in front
        auto partition_point = std::partition(
            particles.begin(), particles.end(),
            [&](const std::unique_ptr<TParticle>& particle) {
                return !particle->isInside(box);
            }
        );
        size_t partition_idx = std::distance(particles.begin(), partition_point);

        // If there are none, stop there
        if (partition_idx == 0) return particles;

        // Otherwise, re-roll a matching number of positions and adjust particle
        // positions accordingly.
        positions = box.pick(rng, partition_idx);
        for (size_t par = 0; par < partition_idx; ++par) {
            particles[par]->m_position = positions[par];
        }
    }
}


void TParticle::propagate(const TDuration& dt) noexcept {
    m_speed += m_force / mass() * dt.count();
    m_position += m_speed * dt.count();
    m_force.setZero();
}


bool TParticle::isInside(const TBox& box) const {
    const std::vector<TVector> normals = box.normalsToInside();
    const std::vector<TVector> projections = box.projectOnSides(m_position);
    for (size_t i = 0; i < normals.size(); ++i) {
        if ((m_position - projections[i]).dot(normals[i]) <= 0.0) return false;
    }
    return true;
}


TVector TParticle::reidInteraction(const TVector& otherPosition,
                                   bool withAttraction) const {
    const TVector distanceVec = m_position - otherPosition;
    const TScalar distance = distanceVec.norm();
    constexpr TScalar mu = 0.7e15;
    const auto forceTerm = [&](int order) {
        return (order * mu * distance - 1)
               / (mu * std::pow(distance, 2))
               * std::exp(-order * mu * distance);
    };
    constexpr TScalar MeV = 1.602177e-13;
    TScalar force = 6484.2 * MeV * forceTerm(7);
    if (withAttraction) {
        force -= 10.463 * MeV * forceTerm(1) + 1650.6 * MeV * forceTerm(4);
    }
    return force * distanceVec / distance;
}


bool TNucleon::isInside(const TBox& box) const {
    const std::vector<TVector> normals = box.normalsToInside();
    const std::vector<TVector> projections = box.projectOnSides(m_position);
    for (size_t i = 0; i < normals.size(); ++i) {
        const TVector edge_projection = projections[i] + s_radius*normals[i];
        if ((m_position - edge_projection).dot(normals[i]) <= 0.0) return false;
    }
    return true;
}


TVector TNucleon::interaction(const TParticle& other) const {
    if (auto other_nucleon = dynamic_cast<const TNucleon*>(&other)) {
        return reidInteraction(other_nucleon->m_position, true);
    } else {
        return TVector::zero(m_position.dims());
    }
}


TVector TChargedParticle::interaction(const TParticle& other) const {
    if (auto other_charged = dynamic_cast<const TChargedParticle*>(&other)) {
        constexpr double K = 8.987551792314e9;  // N.m²/C²
        const TVector distance_vec = m_position - other_charged->m_position;
        const TScalar distance2 = distance_vec.norm2();
        return K * charge() * other_charged->charge() * distance_vec/distance2;
    } else {
        return TVector::zero(m_position.dims());
    }
}


void TPhysics::step() {
    // Force computation
    for (size_t par1 = 0; par1 < m_particles.size(); ++par1) {
        // Particle-box interaction
        m_particles[par1]->interactWith(m_box);

        // Particle-particle interaction
        for (size_t par2 = 0; par2 < m_particles.size(); ++par2) {
            if (par2 == par1) continue;
            m_particles[par1]->interactWith(*m_particles[par2]);
        }
    }

    // Propagation and sanity check
    for (auto& particle: m_particles) {
        // Time propagation
        constexpr TDuration TIME_STEP = std::chrono::nanoseconds(1);
        particle->propagate(TIME_STEP);

        // Particle sanity check
        if (!particle->isInside(m_box)) {
            throw std::runtime_error("A particle ran through the box! \
                Consider decreasing the time step...");
        }
    }
}
