#pragma once

// This code was written specifically for software performance optimization
// trainings. Any similarity with another codebase that you may have seen
// earlier in your career would be a totally unintended coincidence.

#include <cmath>
#include <algorithm>
#include <chrono>
#include <functional>
#include <initializer_list>
#include <memory>
#include <random>
#include <vector>


// Time measurement unit
using TDuration = std::chrono::duration<double>;

// Random number generator
using TRandomEngine = std::mt19937_64;

// Scalar type used for physics computations
using TScalar = double;


// Linear algebra vector
class TVector {
private:
    using Inner = std::vector<TScalar>;

public:
    // Need copy operations for this work
    TVector(const TVector&) noexcept = default;
    TVector& operator=(const TVector&) noexcept = default;

    // Construction from coordinates
    explicit TVector(const std::initializer_list<TScalar>& coords) noexcept
        : m_coords{coords}
    {}
    //
    explicit TVector(const std::vector<TScalar>& coords) noexcept
        : m_coords(coords)
    {}

    // Zero vector for an N-dimensional space
    static TVector zero(size_t num_dims) noexcept {
        return TVector(Inner(num_dims, 0.0));
    }

    // Unit vector for a certain dimension in an N-dimensional space
    // Dimensions are counted from 0.
    static TVector unit(size_t unit_dim, size_t num_dims);

    // Dimension of the vector space
    size_t dims() const noexcept { return m_coords.size(); }

    // Access the i-th component
    TScalar operator[](size_t dim) const;
    TScalar& operator[](size_t dim);

    // Standard linear algebra operators
    TVector operator-() const noexcept {
        return scalarOp([](TScalar coord) { return -coord; });
    }
    //
    TVector operator*(const TScalar x) const noexcept {
        return scalarOp([x](TScalar coord) { return coord * x; });
    }
    //
    friend TVector operator*(const TScalar x, const TVector& v) noexcept {
        return v * x;
    }
    //
    TVector operator/(const TScalar x) const noexcept {
        return scalarOp([x](TScalar coord) { return coord / x; });
    }
    //
    TVector operator+(const TVector& other) const {
        return vectorOp([](TScalar c1, TScalar c2) { return c1 + c2; },
                        other);
    }
    //
    TVector& operator+=(const TVector& other) {
        *this = *this + other;
        return *this;
    }
    //
    TVector operator-(const TVector& other) const {
        return vectorOp([](TScalar c1, TScalar c2) { return c1 - c2; },
                        other);
    }

    // Scalar product
    TScalar dot(const TVector& other) const;

    // L²-norm
    TScalar norm2() const noexcept { return this->dot(*this); }
    TScalar norm() const noexcept { return std::sqrt(norm2()); }

    // Set all components to zero
    void setZero() noexcept {
        for (auto& elem : m_coords) { elem = 0.0; }
    }

    // Assert that all components are strictly positive
    bool isPositive() const noexcept {
        return std::all_of(m_coords.cbegin(), m_coords.cend(),
                           [](TScalar c) { return c > 0.0; });
    }


private:
    Inner m_coords;

    void checkDimInRange(size_t dim) const { checkDimInRange(dim, dims()); }
    static void checkDimInRange(size_t dim, size_t num_dims);

    void checkDimsMatch(const TVector& other) const;

    TVector scalarOp(std::function<TScalar(TScalar)>&& op) const noexcept;

    TVector vectorOp(std::function<TScalar(TScalar, TScalar)>&& op,
                     const TVector& other) const;
};


// A parallelepipedic area in which particles can be put
class TBox {
public:
    // Construct from position and dimensions
    TBox(const TVector& position, const TVector& extent);

    // Box dimensionality
    size_t dims() const noexcept { return m_start.dims(); }

    // Compute a set of normal vectors pointing towards the inside of the box,
    // one per side of the box.
    std::vector<TVector> normalsToInside() const noexcept;

    // Project a particle's position on all sides of the box
    // Sides are enumerated in the same order as in normalsToInside's output.
    std::vector<TVector> projectOnSides(const TVector& position) const;

    // Pick a set of random positions within the box
    std::vector<TVector> pick(TRandomEngine& rng, size_t amount) const noexcept;

private:
    TVector m_start;
    TVector m_end;
};


// A simple Newtonian particle
class TParticle {
public:
    // Need a virtual destructor since we'll use pointer-to-base
    virtual ~TParticle() = default;

    // Generate an idle particle at a certain spatial position
    explicit TParticle(const TVector& position) noexcept
        : m_position(position)
        , m_speed(TVector::zero(position.dims()))
        , m_force(TVector::zero(position.dims()))
    {}

    // Generate multiple particles in a box
    static std::vector<std::unique_ptr<TParticle>> generate(
        TRandomEngine& rng,
        const TBox& box,
        size_t amount
    ) noexcept;

    // Particle mass in kg
    virtual TScalar mass() const noexcept = 0;

    // Simulate interaction with another particle (call with every other particle)
    void interactWith(const TParticle& other) { m_force += interaction(other); }

    // Simulate interaction with TBox boundaries
    void interactWith(const TBox& box) {
        for (const auto& projection: box.projectOnSides(m_position)) {
            m_force += reidInteraction(projection, false);
        }
    }

    // Simulate the passage of time and reset the force
    void propagate(const TDuration& dt) noexcept;

    // Is this particle inside of the TBox ?
    virtual bool isInside(const TBox& box) const;

protected:
    // Compute interaction with another particle
    virtual TVector interaction(const TParticle& other) const = 0;

    // Derive a force from the Reid potential, with the option to remove the
    // attractive terms (used as a toy model of contact interaction)
    TVector reidInteraction(const TVector& otherPosition,
                            bool withAttraction = true) const;

    // Particle position
    TVector m_position;

private:
    // Particle speed and force accumulator
    TVector m_speed;
    TVector m_force;
};


// Nucleons have nuclear interactions with other nucleons
class TNucleon : public virtual TParticle {
public:
    // Override of the TBox inside query that takes radius into account
    bool isInside(const TBox& box) const final override;

protected:
    // Nuclear force
    TVector interaction(const TParticle& other) const override;

private:
    // Nucleon radius
    static constexpr TScalar s_radius = 0.833e-15;
};


// A neutron is a kind of nucleon
class TNeutron final : public TNucleon {
public:
    explicit TNeutron(const TVector& position) noexcept : TParticle(position) {}

    TScalar mass() const noexcept override {
        return 1.6749274980495e-27;
    }
};


// Charged particles have a charge, and thus Coulombian interactions
// (we'll ignore magnetic interaction for simplicity in this toy model)
class TChargedParticle : public virtual TParticle {
public:
    // Particle charge
    virtual TScalar charge() const noexcept = 0;

protected:
    // Coulombian interaction
    TVector interaction(const TParticle& other) const override;

    // Elementary charge
    static constexpr TScalar s_elementaryCharge = 1.602176634e-19;
};


// An electron is a kind of charged particle
class TElectron final : public TChargedParticle {
public:
    explicit TElectron(const TVector& position) noexcept : TParticle(position) {}

    TScalar mass() const noexcept override {
        return 9.109383701528e-31;
    }

    TScalar charge() const noexcept override {
        return -s_elementaryCharge;
    }
};


// A proton is a charged nucleon
class TProton final : public TChargedParticle, public TNucleon {
public:
    explicit TProton(const TVector& position) noexcept : TParticle(position) {}

    TScalar mass() const noexcept override {
        return 1.6726219236951e-27;
    }

    TScalar charge() const noexcept override {
        return s_elementaryCharge;
    }

    TVector interaction(const TParticle& other) const override {
        return TChargedParticle::interaction(other)
               + TNucleon::interaction(other);
    }
};


// Main simulation class
class TPhysics {
public:
    // Set up simulation with a certain number of particles
    explicit TPhysics(TRandomEngine& rng, size_t num_particles) noexcept
        : m_box(TVector::zero(3),
                BOX_SIZE * TVector{ 1.0, 1.0, 1.0 })
        , m_particles(TParticle::generate(rng, m_box, num_particles))
    {}

    // Execute one simulation time step
    void step();

private:
    static constexpr TScalar BOX_SIZE = 0.1;  // m
    TBox m_box;
    std::vector<std::unique_ptr<TParticle>> m_particles;
};
