#pragma once

#include <cstddef>
#include <cstdlib>
#include <limits>
#include <new>
#include <numeric>
#include <type_traits>
#include <vector>

// Force the compiler to assume that any data accessible via the input reference
// may have been read and written. This will cause any register-resident data to
// be flushed to memory, and can be used to make sure that the compiler doesn't
// unfairly optimize simple microbenchmarks.
template <typename T>
inline void assumeAccessed(T&& clobber) {
  __asm__ volatile("" : : "g"(&clobber) : "memory");
}

/// Assume that a value which fits in a CPU register has been read
template <typename T>
inline void assumeUsed(T clobber) {
  __asm__ volatile("" : : "r"(clobber) :);
}

/// STL allocator that produces buffers aligned on 512-bit boundaries
template <typename T>
struct simd_friendly_allocator
{
  using value_type = T;

  static constexpr std::size_t cache_line_size = 512 / 8;
  static constexpr std::size_t alignment = std::lcm(std::alignment_of_v<T>, cache_line_size);

  constexpr simd_friendly_allocator() {}
  template <typename U> simd_friendly_allocator(const simd_friendly_allocator<U>&) {}
  template <typename U> simd_friendly_allocator& operator=(const simd_friendly_allocator<U>&) {}

  bool operator==(const simd_friendly_allocator&) { return true; }
  bool operator!=(const simd_friendly_allocator&) { return false; }

  T* allocate(std::size_t n) {
    if (n > std::numeric_limits<std::size_t>::max() / sizeof(T))
      throw std::bad_array_new_length();

    const std::size_t size = n * sizeof(T);
    const std::size_t blocks =
      size / alignment + static_cast<std::size_t>(size % alignment != 0);
    void* const allocation = std::aligned_alloc(alignment, blocks * alignment);

    if (allocation == nullptr) {
      throw std::bad_alloc();
    }

    return static_cast<T*>(allocation);
  }

  void deallocate(T* p, size_t) {
    std::free(static_cast<void*>(p));
  }
};
//
template <typename T>
using simd_friendly_vector = std::vector<T, simd_friendly_allocator<T>>;

/// STL vector that produces buffers aligned on cache line boundaries.
/// Currently equivalent to simd_friendly_vector on x86 hardware
template <typename T>
using cache_aligned_vector = simd_friendly_vector<T>;
