// This test program writes to a storage device with maximal intensity and can
// be used for the purpose of evaluating the degree of interference between
// storage-bound interactive tasks and batch benchmark jobs.
//
// It takes three CLI parameters:
// - The first parameter indicates whether sequential ("seq") or random ("rand")
//   I/O should be exercised by the benchmark.
// - The second parameter indicates a directory on the target device which the
//   user has write access to.
// - The third parameter indicates the amount of bytes that should be written to
//   the device as a power of 2 (given an input N, 2^N bytes will be written).

#define _GNU_SOURCE

#include <errno.h>
#include <fcntl.h>
#include <omp.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/random.h>
#include <sys/stat.h>

#ifndef _OPENMP
    #error "This benchmark must be built with OpenMP support"
#endif

#define BLOCK_SIZE_POW2 9
#define BLOCK_SIZE (1 << BLOCK_SIZE_POW2)
typedef uint64_t Word;

// === xoroshiro128+ by David Blackman and Sebastiano Vigna ===

static inline uint64_t rotl(const uint64_t x, int k) {
    return (x << k) | (x >> (64 - k));
}

typedef struct {
    uint64_t s[2];
} RNGState;

static inline Word next(RNGState* state) {
    const uint64_t s0 = state->s[0];
    uint64_t s1 = state->s[1];
    const uint64_t result = s0 + s1;

    s1 ^= s0;
    state->s[0] = rotl(s0, 24) ^ s1 ^ (s1 << 16); // a, b
    state->s[1] = rotl(s1, 37); // c

    return result;
}

// ======

typedef enum { SEQUENTIAL, RANDOM } Mode;

int main(int argc, char *argv[]) {
    // Get parameters from user CLI input
    if (argc != 4) {
        printf("Please specify...\n"
               "- Whether I/O should be sequential (\"seq\") or random (\"rand\")\n"
               "- A directory representing the target device\n"
               "- The base-2 log of the number of bytes to be written\n");
        return 1;
    }
    //
    Mode mode;
    if (strcmp(argv[1], "seq") == 0) {
        mode = SEQUENTIAL;
    } else if (strcmp(argv[1], "rand") == 0) {
        mode = RANDOM;
    } else {
        printf("Invalid operating mode: \"%s\"\n", argv[1]);
        return 1;
    }
    //
    const char* const target_path = argv[2];
    //
    errno = 0;
    const uint64_t num_bytes_pow2 = strtoul(argv[3], NULL, 10);
    if ((errno != 0) || (num_bytes_pow2 < BLOCK_SIZE_POW2) || (num_bytes_pow2 >= 64)) {
        printf("Error: \"%s\" is not a valid input\n", argv[1]);
        return 1;
    }
    uint64_t num_bytes = 1ULL << num_bytes_pow2;

    // Allocate a buffer of data to be filled
    Word* buffer;
    int result = posix_memalign((void**)&buffer, BLOCK_SIZE, num_bytes);
    if (result != 0) {
        printf("Failed to allocate aligned buffer\n");
        return 1;
    }
    buffer = __builtin_assume_aligned(buffer, BLOCK_SIZE);

    // Fill the buffer with random data in parallel
    //
    // Random data must be used because some storage devices implement
    // compression tricks that make the I/O rate pattern-dependent, and in
    // particular unrealistically fast for the classic all-zeroes pattern...
    //
    result = 0;
    #pragma omp parallel
    {
        // Set up a thread-local random number generator
        RNGState state;
        if (getrandom(&state, sizeof(state), 0) != sizeof(state)) {
            printf("Failed to initialize RNG state\n");
            result = 1;
            #pragma omp cancel parallel
        }

        // Generate the random data
        #pragma omp for
        for (uint64_t word = 0; word < num_bytes / sizeof(Word); ++word) {
            buffer[word] = next(&state);
        }
    }
    if (result != 0) return result;

    // Open a temporary file to write the data on the device, and use direct I/O
    // so that the Linux disk cache doesn't bias the benchmark.
    int fd = open(target_path,
                  O_DIRECT | O_TMPFILE | O_WRONLY,
                  S_IWUSR);
    if (fd < 0) {
        printf("Failed to open storage with errno %d\n", errno);
        return 1;
    }

    // Give the OS advance notice of how long the file will be
    if (posix_fallocate(fd, 0, num_bytes) < 0) {
        printf("Failed to allocate storage with errno %d\n", errno);
        return 1;
    }

    // Write the data to the device
    const char* current_byte = (const char*)buffer;
    switch (mode) {
        case SEQUENTIAL:
            while (num_bytes) {
                const ssize_t written = write(fd, current_byte, num_bytes);
                if (written <= 0) {
                    printf("Failed to write data to storage with errno %d\n", errno);
                    return 1;
                }
                num_bytes -= written;
                current_byte += written;
            }
            break;
        case RANDOM:
            while (num_bytes) {
                const uint64_t randomness = *((Word*)current_byte);
                uint64_t position = (randomness % num_bytes) / BLOCK_SIZE * BLOCK_SIZE;
                int remaining = BLOCK_SIZE;
                while (remaining) {
                    ssize_t written = pwrite(fd, current_byte, remaining, position);
                    if (written <= 0) {
                        printf("Failed to write data to storage with errno %d\n", errno);
                        return 1;
                    }
                    current_byte += written;
                    position += written;
                    remaining -= written;
                }
                num_bytes -= BLOCK_SIZE;
            }
            break;
        default:
            printf("Unexpected operating mode\n");
            return 1;
    }

    // Wait for data to be written to the device before quitting
    if (fsync(fd) < 0) {
        printf("Failed to commit data writes with errno %d\n", errno);
        return 1;
    }

    return 0;
}
